"""Define the schema of the package definition file."""

from typing import Dict

import jsonschema

from aws_orga_deployer import config
from aws_orga_deployer.engines import ENGINES


def validate(content: Dict) -> None:
    """Validate the content provided with the schema.

    Args:
        content: Dict to validate.

    Raise:
        ValidationError: If the schema is invalid.
    """
    schema = {
        "type": "object",
        "additionalProperties": True,
        "properties": {
            "PackageConfiguration": {
                "description": "Package configuration settings",
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "S3Bucket": {
                        "description": "Name of the S3 bucket to store persistent data",
                        "type": "string",
                        "minLength": 3,
                    },
                    "S3Region": {
                        "description": "Region of the S3 bucket",
                        "type": "string",
                        "minLength": 3,
                    },
                    "S3Prefix": {
                        "description": "S3 prefix. Must ends with a slash if specified",
                        "type": "string",
                        "pattern": "^.+\\/$",
                    },
                    "ConcurrentWorkers": {
                        "description": "Number of concurrent threads deploying modules",
                        "type": "number",
                        "minimum": 1,
                        "maximum": 50,
                    },
                    "AssumeOrgaRoleArn": {
                        "description": (
                            "ARN of the IAM role to assume to query AWS Organizations"
                        ),
                        "type": "string",
                    },
                    "OrgaCacheExpiration": {
                        "description": (
                            "Period in seconds during which the cache in S3 of"
                            " information on accounts and organizational units is"
                            " reused instead of querying AWS Organizations"
                        ),
                        "type": "number",
                        "minimum": 0,
                    },
                    "OverrideAccountNameByTag": {
                        "description": (
                            "Tag key assigned to AWS accounts whose value is used to"
                            " override the account name"
                        ),
                        "type": "string",
                    },
                },
                "required": ["S3Bucket", "S3Region"],
            },
            "DefaultModuleConfiguration": {
                "description": (
                    "Default configuration settings for all modules or modules of a"
                    " given engine"
                ),
                "type": "object",
                "propertyNames": {"enum": ["All", *ENGINES]},
                "patternProperties": {"^": {"type": "object"}},
                "additionalProperties": False,
            },
            "DefaultVariables": {
                "description": (
                    "Default variables for all modules or modules of a given engine"
                ),
                "type": "object",
                "propertyNames": {"enum": ["All", *ENGINES]},
                "patternProperties": {"^": {"type": "object"}},
                "additionalProperties": False,
            },
            "Modules": {
                "description": "Module deployments definition",
                "type": "object",
                "propertyNames": {"enum": [*config.MODULES]},
                "patternProperties": {
                    "^": {
                        "type": "object",
                        "additionalProperties": False,
                        "properties": {
                            "Configuration": {
                                "description": "Module configuration settings",
                                "type": "object",
                            },
                            "Variables": {
                                "description": "Module variables",
                                "type": "object",
                            },
                            "VariablesFromOutputs": {
                                "type": "object",
                                "patternProperties": {
                                    "^": {
                                        "type": "object",
                                        "additionalProperties": False,
                                        "properties": {
                                            "Module": {
                                                "type": "string",
                                                "enum": [*config.MODULES],
                                            },
                                            "Region": {"type": "string"},
                                            "AccountId": {"type": "string"},
                                            "OutputName": {"type": "string"},
                                        },
                                        "required": [
                                            "Module",
                                            "Region",
                                            "AccountId",
                                            "OutputName",
                                        ],
                                    }
                                },
                            },
                            "Deployments": {
                                "description": "List of module deployments",
                                "type": "array",
                                "items": {
                                    "type": "object",
                                    "additionalProperties": False,
                                    "properties": {
                                        "Include": {
                                            "type": "object",
                                            "additionalProperties": False,
                                            "properties": {
                                                "AccountIds": {
                                                    "type": "array",
                                                    "items": {
                                                        "type": "string",
                                                        "pattern": "^[0-9]{12}$",
                                                    },
                                                },
                                                "AccountNames": {
                                                    "type": "array",
                                                    "items": {"type": "string"},
                                                },
                                                "AccountTags": {
                                                    "type": "array",
                                                    "items": {
                                                        "type": "string",
                                                        "pattern": "^.+=.+$",
                                                    },
                                                },
                                                "OUIds": {
                                                    "type": "array",
                                                    "items": {"type": "string"},
                                                },
                                                "OUTags": {
                                                    "type": "array",
                                                    "items": {
                                                        "type": "string",
                                                        "pattern": "^.+=.+$",
                                                    },
                                                },
                                                "Regions": {
                                                    "type": "array",
                                                    "items": {"type": "string"},
                                                },
                                            },
                                        },
                                        "Exclude": {
                                            "type": "object",
                                            "additionalProperties": False,
                                            "properties": {
                                                "AccountIds": {
                                                    "type": "array",
                                                    "items": {
                                                        "type": "string",
                                                        "pattern": "^[0-9]{12}$",
                                                    },
                                                },
                                                "AccountNames": {
                                                    "type": "array",
                                                    "items": {"type": "string"},
                                                },
                                                "AccountTags": {
                                                    "type": "array",
                                                    "items": {
                                                        "type": "string",
                                                        "pattern": "^.+=.+$",
                                                    },
                                                },
                                                "OUIds": {
                                                    "type": "array",
                                                    "items": {"type": "string"},
                                                },
                                                "OUTags": {
                                                    "type": "array",
                                                    "items": {
                                                        "type": "string",
                                                        "pattern": "^.+=.+$",
                                                    },
                                                },
                                                "Regions": {
                                                    "type": "array",
                                                    "items": {"type": "string"},
                                                },
                                            },
                                        },
                                        "Variables": {"type": "object"},
                                        "VariablesFromOutputs": {
                                            "type": "object",
                                            "patternProperties": {
                                                "^": {
                                                    "type": "object",
                                                    "additionalProperties": False,
                                                    "properties": {
                                                        "Module": {
                                                            "type": "string",
                                                            "enum": [*config.MODULES],
                                                        },
                                                        "Region": {"type": "string"},
                                                        "AccountId": {"type": "string"},
                                                        "OutputName": {
                                                            "type": "string"
                                                        },
                                                        "IgnoreIfNotExists": {
                                                            "type": "boolean"
                                                        },
                                                    },
                                                    "required": [
                                                        "Module",
                                                        "Region",
                                                        "AccountId",
                                                        "OutputName",
                                                    ],
                                                }
                                            },
                                        },
                                        "Dependencies": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "additionalProperties": False,
                                                "properties": {
                                                    "Module": {
                                                        "type": "string",
                                                        "enum": [*config.MODULES],
                                                    },
                                                    "Region": {"type": "string"},
                                                    "AccountId": {"type": "string"},
                                                    "IgnoreIfNotExists": {
                                                        "type": "boolean"
                                                    },
                                                },
                                                "required": [
                                                    "Module",
                                                    "Region",
                                                    "AccountId",
                                                ],
                                            },
                                        },
                                    },
                                },
                            },
                        },
                        "required": ["Deployments"],
                    }
                },
            },
        },
        "required": ["PackageConfiguration", "Modules"],
    }
    jsonschema.validate(instance=content, schema=schema)



"""General utility functions."""

# COMPLETED
import json
import logging
import queue
from threading import Thread
from typing import Any, Callable, Dict, List, Optional

import boto3
from botocore.config import Config
from botocore.session import Session

from aws_orga_deployer import config

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


def get_aws_temp_credentials(
    iam_role: str,
    sts_session: Optional[Session] = None,
) -> Dict[str, str]:
    """Assume an IAM role and return AWS temporary credentials.

    Args:
        iam_role: ARN of the IAM role to assume.
        sts_session: boto3 Session to use for the STS call. Defaut is None to
            use a new session with current AWS credentials.

    Returns:
        dict: With `AccessKeyId`, `SecretAccessKey` and `SessionToken`
            attributes.
    """
    if sts_session is None:
        sts_client = boto3.client("sts")
    else:
        sts_client = sts_session.client("sts")
    LOGGER.debug("Assuming the IAM role %s", iam_role)
    response = sts_client.assume_role(
        RoleArn=iam_role, RoleSessionName="aws-orga-deployer"
    )
    return response["Credentials"]


def get_aws_session(
    iam_role: Optional[str] = None,
    sts_session: Optional[Session] = None,
) -> Session:
    """Assume an IAM role if needed and return a boto3 Session object.

    Args:
        iam_role: ARN of the IAM role to assume. Defaut is None to use current
            AWS credentials.
        sts_session: boto3 Session to use for the STS call. Defaut is None to
            use a new session with current AWS credentials.

    Returns:
        Session: boto3 Session.
    """
    # If a role must be assumed
    if iam_role:
        credentials = get_aws_temp_credentials(iam_role, sts_session)
        return boto3.session.Session(
            aws_access_key_id=credentials["AccessKeyId"],
            aws_secret_access_key=credentials["SecretAccessKey"],
            aws_session_token=credentials["SessionToken"],
        )
    # If current AWS credentials are used
    return boto3.session.Session()


def get_aws_client(session: Session, *args: Any, **kwargs: Any) -> Any:
    """Return a boto3 Client with retry configuration.

    Args:
        session: boto3 Session to use.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        boto3 Client object.
    """
    boto_config = Config(retries={"max_attempts": 10, "mode": "standard"})
    return session.client(*args, config=boto_config, **kwargs)


class PropagatingThread(Thread):
    """Thread object that raises exception in the parent thread."""

    # pylint: disable=attribute-defined-outside-init
    # pylint: disable=broad-exception-caught

    def run(self) -> None:
        self.exc = None
        try:
            self.ret = self._target(*self._args, **self._kwargs)
        except BaseException as exception:
            self.exc = exception

    def join(self, timeout: Optional[float] = None) -> Any:
        super().join(timeout)
        if self.exc:
            raise self.exc
        return self.ret


def exec_multithread(items: List[Any], func: Callable, concurrency: int) -> None:
    """Execute a function for each item of a list using concurrent threads.

    Args:
        items: List of items to process.
        func: Function to use to process each item.
        concurrency: Number of concurrent threads.
    """
    items_queue: queue.Queue = queue.Queue()
    # Populate the queue
    for item in items:
        items_queue.put(item)

    # Create a wrapper function that pulls the queue
    def wrapper():
        while True:
            try:
                item = items_queue.get(block=False)
                func(item)
            except queue.Empty:
                break
            items_queue.task_done()

    # Launch concurrent threads
    threads = []
    for _ in range(concurrency):
        thread = PropagatingThread(target=wrapper, daemon=True)
        thread.start()
        threads.append(thread)
    # Wait until the threads complete
    for thread in threads:
        thread.join()


def get_s3_key(object_path: str) -> str:
    """Prepends the S3 prefix to the object path and returns the result.

    Args:
        object_path: Object path

    Returns:
        S3Prefix concatenated with object path.
    """
    return config.PACKAGE.get("S3Prefix", "") + object_path


class FileNotExists(Exception):
    """Exception raised if the S3 object does not exist when using
    `load_dict_from_s3`.
    """


def load_json_from_s3(object_path: str) -> Dict:
    """Load a JSON file from S3 and return a dict.

    Args:
        object_path: Object path in the dedicated S3 prefix.

    Returns:
        Dictionary containing the JSON file.

    Raises:
        FileNotExists: If the object does not exist in S3.
    """
    bucket = config.PACKAGE["S3Bucket"]
    key = get_s3_key(object_path)
    client = get_aws_client(
        get_aws_session(), "s3", region_name=config.PACKAGE["S3Region"]
    )
    try:
        LOGGER.debug("Reading the S3 object at s3://%s/%s", bucket, key)
        response = client.get_object(Bucket=bucket, Key=key)
        content_bytes = response["Body"].read()
        return json.loads(content_bytes.decode())
    except client.exceptions.NoSuchKey as err:
        raise FileNotExists from err


def write_dict_to_s3(content: Dict, object_path: str) -> None:
    """Write a dict to Amazon S3 as a JSON file.

    Args:
        content: Content to wroite.
        object_path: Object path in the dedicated S3 prefix.
    """
    content_bytes = json.dumps(content, indent=4).encode()
    bucket = config.PACKAGE["S3Bucket"]
    key = get_s3_key(object_path)
    client = get_aws_client(
        get_aws_session(), "s3", region_name=config.PACKAGE["S3Region"]
    )
    LOGGER.debug("Writing the S3 object at s3://%s/%s", bucket, key)
    client.put_object(
        Body=content_bytes, Bucket=bucket, Key=key, ContentType="application/json"
    )


def write_output_json(input_dict: Dict, description: str) -> None:
    """Write the input dictionary as a JSON file to the local disk.

    Args:
        input_dict: Input content
        description: Description of the file to print in the logs
    """
    LOGGER.info("Exporting %s to %s", description, config.CLI["output_file"])
    with open(config.CLI["output_file"], "w", encoding="utf-8") as stream:
        content = json.dumps(input_dict, indent=4)
        stream.write(content)



"""Test the module `aws_orga_deployer.modules`."""

# COMPLETED
import unittest

from aws_orga_deployer import config, modules
from tests import mock


class TestModules(unittest.TestCase):
    """Test the modules."""

    def setUp(self):
        """Load the modules."""
        mock.mock_cli_arguments(package_filename="package1.yaml")
        modules.load_modules()

    def test_modules(self):
        """Check that the modules are loaded."""
        self.assertIn("python1", config.MODULES)
        self.assertEqual(config.MODULES["python1"].engine, "python")

    def test_module_hash(self):
        """Check that the module "python1" has a module hash."""
        module_hash = config.MODULES["python1"].module_hash
        self.assertRegex(module_hash, "^[a-f0-9]{32}$")

    def test_patterns(self):
        """Check that the hash configuration file is correctly used."""
        self.assertEqual(
            config.MODULES["python1"].included_patterns,
            ["*.py"],
        )
        self.assertEqual(
            config.MODULES["python1"].excluded_patterns,
            ["~*.py"],
        )

    def test_module_config_python_invalid_1(self):
        """Check an invalid module configuration for Python modules."""
        module_config = {
            "AssumeRole": None,
            "Retry": {"MaxAttempts": 1, "DelayBeforeRetrying": 0},
            "PythonExecutable": [],
        }
        with self.assertRaises(AssertionError):
            config.MODULES["python1"].validate_module_config(module_config)

    def test_module_config_python_invalid_2(self):
        """Check an invalid module configuration for Python modules."""
        module_config = {"Retry": {"MaxAttempts": 0, "DelayBeforeRetrying": 0}}
        with self.assertRaises(AssertionError):
            config.MODULES["python1"].validate_module_config(module_config)

    def test_module_config_python_valid(self):
        """Check an valid module configuration for Python modules."""
        module_config = {}
        config.MODULES["python1"].validate_module_config(module_config)

    def test_module_config_cloudformation_invalid_1(self):
        """Check an invalid module configuration for CloudFormation modules."""
        module_config = {}
        with self.assertRaises(AssertionError):
            config.MODULES["cloudformation1"].validate_module_config(module_config)

    def test_module_config_cloudformation_invalid_2(self):
        """Check an invalid module configuration for CloudFormation modules."""
        module_config = {"StackName": "test"}
        with self.assertRaises(AssertionError):
            config.MODULES["cloudformation1"].validate_module_config(module_config)

    def test_module_config_cloudformation_valid(self):
        """Check an invalid module configuration for CloudFormation modules."""
        module_config = {"StackName": "test", "TemplateFilename": "template.yaml"}
        config.MODULES["cloudformation1"].validate_module_config(module_config)

    def test_module_config_terraform_invalid(self):
        """Check an invalid module configuration for CloudFormation modules."""
        module_config = {"TerraformExecutable": 1}
        with self.assertRaises(AssertionError):
            config.MODULES["terraform1"].validate_module_config(module_config)

    def test_module_config_terraform_valid(self):
        """Check an invalid module configuration for CloudFormation modules."""
        module_config = {"TerraformExecutable": "/bin/terraform"}
        config.MODULES["terraform1"].validate_module_config(module_config)



"""Information on AWS accounts and organizational units."""

# COMPLETED
import logging
import re
from datetime import datetime
from typing import Any, Dict, List, Optional, Set

import botocore

from aws_orga_deployer import config, utils

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


class OrgaParser:
    """Class to retrieve and use information from AWS Organizations.

    Attributes:
        accounts: Information about accounts
        ous: Information about organizational units
        loaded_from_cache: True if it was loaded from the cache in S3
    """

    accounts: Dict[str, Any]
    ous: Dict[str, Any]
    loaded_from_cache: bool

    def __init__(self) -> None:
        """Load and parse information on AWS accounts and OUs."""
        # If the information on accounts and organizational units could not be
        # loaded from the cache in S3, query AWS Organizations
        self.loaded_from_cache = self._load_from_cache()
        if self.loaded_from_cache is False:
            LOGGER.info(
                "Querying AWS Organizations for information on accounts and"
                " organizational units"
            )
            # Create a session with permissions to query AWS Organizations
            self.master_session = utils.get_aws_session(
                config.PACKAGE.get("AssumeOrgaRoleArn")
            )
            # Query AWS Organizations
            self._fetch_accounts()
            self._fetch_ous()
            # Update the cache in S3
            content = self.export()
            utils.write_dict_to_s3(content, config.ORGA_CACHE_FILENAME)
        LOGGER.info(
            "Found %s accounts and %s organizational units",
            len(self.accounts),
            len(self.ous),
        )

    def _load_from_cache(self) -> bool:
        """Load information on accounts and organizational units from the cache
        in S3 if it exists and has not expired, and if the CLI argument
        `--force-orga-refresh` is not passed.

        Returns:
            True if the cache in S3 was used to load information about accounts
                and organizational units, False otherwise.
        """
        # If the argument --force-orga-refresh is passed, return False
        if config.CLI.get("force_orga_refresh", False):
            LOGGER.debug(
                "Ignoring the cache in S3 and forcing the tool to query AWS"
                " Organizations"
            )
            return False
        client = utils.get_aws_client(
            utils.get_aws_session(), "s3", region_name=config.PACKAGE["S3Region"]
        )
        bucket = config.PACKAGE["S3Bucket"]
        key = utils.get_s3_key(config.ORGA_CACHE_FILENAME)
        # Returns the number of seconds since the cache in S3 was modified for
        # the last time
        try:
            LOGGER.debug(
                "Retrieving the last modification date of the cache of information"
                " on accounts and organizational units at s3://%s/%s",
                bucket,
                key,
            )
            response = client.head_object(Bucket=bucket, Key=key)
            modified_date = response["LastModified"].replace(tzinfo=None)
            seconds = (datetime.utcnow() - modified_date).total_seconds()
            expiration = config.PACKAGE.get(
                "OrgaCacheExpiration",
                config.DEFAULT_ORGA_CACHE_EXPIRATION,
            )
            # Load information from the cache if it has not expired
            if seconds <= expiration:
                LOGGER.info(
                    "Loading information on accounts and organizational units from the"
                    " cache in S3"
                )
                cache = utils.load_json_from_s3(config.ORGA_CACHE_FILENAME)
                self.accounts = cache["Accounts"]
                self.ous = cache["OUs"]
                return True
            # If the cache has expired
            LOGGER.debug(
                "The cache in S3 of information on accounts and organizational units"
                " has expired"
            )
            return False
        # Return False if there is no cache in S3 or if the cache could not be
        # loaded
        except botocore.exceptions.ClientError as err:
            if (
                isinstance(err, botocore.exceptions.ClientError)
                and err.response["Error"]["Message"] == "Not Found"
            ):
                LOGGER.debug(
                    "There is no cache of information on accounts and organizational"
                    " units in S3"
                )
            else:
                LOGGER.exception(
                    "Failed to load the cache of information on accounts and"
                    " organizational units from S3",
                    exc_info=config.CLI["debug"],
                )
            return False

    def _fetch_accounts(self) -> None:
        """
        Query AWS Organizations for information on AWS accounts and store it in
        the attribute `accounts`.
        """
        orga_client = utils.get_aws_client(self.master_session, "organizations")
        account_client = utils.get_aws_client(self.master_session, "account")
        sts_client = utils.get_aws_client(self.master_session, "sts")
        master_account_id = sts_client.get_caller_identity()["Account"]
        accounts: Dict[str, Dict] = {}

        def browse_ou(ou_id: str, parent_ou_ids: Optional[List[str]] = None) -> None:
            """Retrieve the children of an OU and process recurvisely the
            child OUs.

            Args:
                ou_id: ID of the organizational unit to process.
                parent_ou_ids: List of parent OU IDs. Default corresponds to an
                    empty list.
            """
            ou_ids = [ou_id]
            if not parent_ou_ids is None:
                ou_ids += parent_ou_ids
            # Append the list of parent OUs of each account in this OU
            children_accounts = orga_client.list_children(
                ParentId=ou_id, ChildType="ACCOUNT"
            )["Children"]
            for children_account in children_accounts:
                if children_account["Id"] in accounts:
                    accounts[children_account["Id"]]["ParentOUs"] = ou_ids
            # Browse recursively the organization
            children_ous = orga_client.list_children(
                ParentId=ou_id, ChildType="ORGANIZATIONAL_UNIT"
            )["Children"]
            for children_ou in children_ous:
                browse_ou(children_ou["Id"], ou_ids)

        def get_account_tags_regions(account_id: str) -> None:
            """Retrieve the tags and enabled regions of an account. Update the
            account name if it must be replaced by the value of a given tag if
            it exists.

            Args:
                account_id: ID of the account.
            """
            accounts[account_id]["Tags"] = {}
            # Retrieve the tags
            paginator = orga_client.get_paginator("list_tags_for_resource")
            pages = paginator.paginate(ResourceId=account_id)
            for page in pages:
                for tag in page["Tags"]:
                    accounts[account_id]["Tags"][tag["Key"]] = tag["Value"]
            # Update the account name if needed
            tag_key = config.PACKAGE.get("OverrideAccountNameByTag")
            if tag_key in accounts[account_id]["Tags"]:
                new_account_name = accounts[account_id]["Tags"][tag_key]
                accounts[account_id]["Name"] = new_account_name
            # Retrieve the enabled regions. The argument `AcountId` must not
            # be passed to the `list_regions` request for the master account
            if account_id == master_account_id:
                regions = account_client.list_regions(
                    RegionOptStatusContains=["ENABLED", "ENABLED_BY_DEFAULT"]
                )["Regions"]
            else:
                regions = account_client.list_regions(
                    AccountId=account_id,
                    RegionOptStatusContains=["ENABLED", "ENABLED_BY_DEFAULT"],
                )["Regions"]
            enabled_regions = [region["RegionName"] for region in regions]
            accounts[account_id]["EnabledRegions"] = enabled_regions

        # List the accounts that exist in the organization in a dict with
        # {account_id: {'Name': account_name}}. Exclude non-active accounts.
        paginator = orga_client.get_paginator("list_accounts")
        pages = paginator.paginate()
        for page in pages:
            for account in page["Accounts"]:
                if account["Status"] == "ACTIVE":
                    accounts[account["Id"]] = {"Name": account["Name"]}

        # Recursively browse the organization to find the parent OUs for each
        # account
        roots = orga_client.list_roots()["Roots"]
        for root in roots:
            browse_ou(root["Id"])

        # Retrieve the account tags and enabled the regions for each account
        # using concurrent threads
        utils.exec_multithread(
            list(accounts.keys()),
            get_account_tags_regions,
            config.CONCURRENT_ORGA_THREADS,
        )
        self.accounts = accounts

    def _fetch_ous(self) -> None:
        """
        Query AWS Organizations for information on organizational units and
        store it in the attribute `_ous`.
        """
        orga_client = utils.get_aws_client(self.master_session, "organizations")
        ous: Dict[str, Dict] = {}

        # Get the list of unique OUs
        ou_ids = []
        for account in self.accounts.values():
            ou_ids += account["ParentOUs"]
        ou_ids = list(set(ou_ids))

        def get_ou_name_tags(ou_id: str) -> None:
            """Retrieve the name and the tags of an organizational unit.

            Args:
                ou_id (str): ID of the OU
            """
            ous[ou_id] = {}
            ous[ou_id]["Tags"] = {}
            # Retrieve the name
            if ou_id.startswith("ou-"):
                ous[ou_id]["Name"] = orga_client.describe_organizational_unit(
                    OrganizationalUnitId=ou_id
                )["OrganizationalUnit"]["Name"]
            else:
                ous[ou_id]["Name"] = "root"
            # Retrieve the tags
            paginator = orga_client.get_paginator("list_tags_for_resource")
            pages = paginator.paginate(ResourceId=ou_id)
            for page in pages:
                for tag in page["Tags"]:
                    ous[ou_id]["Tags"][tag["Key"]] = tag["Value"]

        # Retrieve the name and the tags for each OU using concurrent threads
        utils.exec_multithread(ou_ids, get_ou_name_tags, config.CONCURRENT_ORGA_THREADS)
        self.ous = ous

    def export(self) -> Dict[str, Dict[str, Dict]]:
        """Returns a dict with information on AWS accounts and organizational
        units.

        Returns:
            Dictionary whose structure is::

                {
                    'Accounts': accounts,
                    'OUs: ous
                }

        """
        return {"Accounts": self.accounts, "OUs": self.ous}

    def get_all_accounts(self) -> List[str]:
        """Return the list of all accounts IDs in the organization.

        Returns:
            List of account IDs.
        """
        return [*self.accounts]

    def get_accounts_by_id(self, account_ids: List[str]) -> List[str]:
        """Return the list of accounts IDs in the organization that intersect
        with a given list of account IDs.

        Args:
            account_ids: List of account IDs.

        Returns:
            List of account IDs.
        """
        return [account_id for account_id in self.accounts if account_id in account_ids]

    def get_accounts_by_name(self, patterns: List[str]) -> List[str]:
        """Return the list of accounts IDs in the organization whose account
        name matches one of the given patterns.

        Args:
            patterns: List of account name patterns.

        Returns:
            List of account IDs.
        """

        result = []
        for account_id, account in self.accounts.items():
            for pattern in patterns:
                expr = re.escape(pattern).replace("\\*", ".*")
                expr = "^" + expr + "$"
                if re.match(expr, account["Name"]):
                    result.append(account_id)
                    break
        return result

    def get_accounts_by_ou(self, ou_ids: List[str]) -> List[str]:
        """Return the list of account IDs in the organization that belong to
        at least one of the organization units in a given list.

        Args:
            out_ids: List of organizational unit IDs.

        Returns:
            List of account IDs.
        """
        result = []
        for account_id, account in self.accounts.items():
            for parent_ou in account["ParentOUs"]:
                if parent_ou in ou_ids:
                    result.append(account_id)
                    break
        return result

    def get_accounts_by_tag(self, tags: List[str]) -> List[str]:
        """Return the list of account IDs in the organization that have a set of
        given tags assigned.

        Args:
            tags: List of tags TAG_KEY=TAG_VALUE.

        Returns:
            List of account IDs.
        """
        result = []
        for account_id, account in self.accounts.items():
            account_has_all_tags = True
            for tag in tags:
                tag_key, tag_value = tag.split("=")
                if not account["Tags"].get(tag_key) == tag_value:
                    account_has_all_tags = False
                    break
            if account_has_all_tags:
                result.append(account_id)
        return result

    def get_accounts_by_ou_tag(self, tags: List[str]) -> List[str]:
        """Return the list of account IDs in the organization that belong to
        at least one organizational unit that have a set of given tags assigned.

        Args:
            tags: List of tags TAG_KEY=TAG_VALUE.

        Returns:
            List of account IDs.
        """
        result = []
        for account_id, account in self.accounts.items():
            for parent_ou_id in account["ParentOUs"]:
                ou_has_all_tags = True
                parent_ou_tags = self.ous[parent_ou_id]["Tags"]
                for tag in tags:
                    tag_key, tag_value = tag.split("=")
                    if not parent_ou_tags.get(tag_key) == tag_value:
                        ou_has_all_tags = False
                        break
                if ou_has_all_tags:
                    result.append(account_id)
                    break
        return result

    def get_account_regions(self, account_id: str, regions: List[str]) -> List[str]:
        """Return the list of regions that are enabled in a given account and
        that intersect with a given list of regions. `ALL_ENABLED` returns all
        enabled regions.

        Args:
            account_id: Account ID.
            regions: List of regions.

        Returns:
            List of regions.
        """
        if "ALL_ENABLED" in regions:
            return self.accounts[account_id]["EnabledRegions"]
        return [
            region
            for region in self.accounts[account_id]["EnabledRegions"]
            if region in regions
        ]

    def get_all_enabled_regions(self) -> List[str]:
        """Return a list of regions that are enabled in at least one account.

        Return:
            List of regions.
        """
        regions: Set[str] = set()
        for account in self.accounts.values():
            regions = regions | set(account["EnabledRegions"])
        return list(regions)

    def get_account_name(self, account_id: str) -> str:
        """Return the account name of a given account ID.

        Args:
            account_id: Account ID

        Returns:
            Account name
        """
        # The account ID may not exist in `self.accounts` a deployment was made
        # to an account that is not active anymore
        if not account_id in self.accounts:
            return "undefined"
        return self.accounts[account_id]["Name"]

    def account_region_exists(self, account_id: str, region: str) -> bool:
        """Return True if the account exists in the organization and the region
        is enabled in this account.

        Args:
            account_id: Account ID
            region: Region

        Returns:
            True if the account exists and the region is enabled.
        """
        if account_id in self.accounts:
            if region in self.accounts[account_id]["EnabledRegions"]:
                return True
        return False



"""Base functions for wrapper scripts."""

import json
from dataclasses import dataclass
from typing import Any, Dict, Optional


@dataclass
class WrapperInputs:
    """Store inputs that are read from the JSON file `input.json`.

    Attributes:
        module: Deployment module.
        account_id: Deployment AWS account ID.
        region: Deployment region.
        command: CLI command requested ("preview" or "deploy").
        action: Action to be made ("create", "update" or "destroy").
        variables: Input variables.
        module_config: Module configuration.
        module_dir: Path to the module directory.
        deployment_cache_dir: Path to the module deployment cache directory.
        engine_cache_dir: Path to the engine cache directory (common to all
            modules for this engine).
    """

    module: str
    account_id: str
    region: str
    command: str
    action: str
    variables: Dict[str, Any]
    module_config: Dict[str, Any]
    module_dir: str
    deployment_cache_dir: str
    engine_cache_dir: str


def read_wrapper_inputs() -> WrapperInputs:
    """Read the inputs to the wrapper function from the file `input.json`.

    Returns:
        WrapperInputs object.
    """
    with open("input.json", "r", encoding="utf-8") as stream:
        content = json.load(stream)
        return WrapperInputs(
            module=content["Deployment"]["Module"],
            account_id=content["Deployment"]["AccountId"],
            region=content["Deployment"]["Region"],
            command=content["Command"],
            action=content["Action"],
            variables=content["Variables"],
            module_config=content["ModuleConfig"],
            module_dir=content["ModulePath"],
            deployment_cache_dir=content["DeploymentCacheDir"],
            engine_cache_dir=content["EngineCacheDir"],
        )


def write_wrapper_outputs(
    made_changes: bool,
    result: str,
    detailed_results: Optional[Dict[str, Any]] = None,
    outputs: Optional[Dict[str, Any]] = None,
) -> None:
    """Write the outcomes of the wrapper function to a file `output.json`.

    Args:
        made_changes: True if the step resulted in changes made or to be made.
        result: Summary of the result.
        detailed_results: Optional detailed results, such as the list of
            resources added, changed or deleted. Should only be valued for
            completed steps.
        outputs: Outputs returned when the step has completed. Should be valued
            only when the command is "apply" and if the action is not "destroy".
    """
    output_content = {
        "MadeChanges": made_changes,
        "Result": result,
        "DetailedResults": detailed_results,
        "Outputs": outputs,
    }
    with open("output.json", "w", encoding="utf-8") as stream:
        json.dump(output_content, stream)



"""Test the sub-package `package.engines`."""

# COMPLETED
from aws_orga_deployer.engines import ENGINES


def test_list_engines():
    """Check that the engines for python, terraform and cloudformation are
    successfully loaded.
    """
    for engine in ("python", "terraform", "cloudformation"):
        assert engine in ENGINES



"""Load modules."""

# COMPLETED
import logging
import os
from typing import Dict

from aws_orga_deployer import config
from aws_orga_deployer.engines import ENGINES
from aws_orga_deployer.engines.base import BaseEngine

LOGGER = logging.getLogger(__name__)  # pylint: disable=invalid-name
LOGGER.addHandler(logging.NullHandler())


class ModuleError(Exception):
    """Exception raised when the package contains invalid modules."""


def load_modules() -> None:
    """Instantiate the Engine class for each module and store them in a
    dict in `config.MODULES`.
    """
    modules: Dict[str, BaseEngine] = {}
    modules_stats: Dict[str, int] = {}
    # The modules must be in the same folder than the package file. The
    # first level of subfolders correspond to the engine, the second to the
    # modules
    package_path = os.path.dirname(os.path.abspath(config.CLI["package_file"]))
    # For each type of engine
    for engine, engine_class in ENGINES.items():
        engine_path = os.path.join(package_path, engine)
        # If there is a folder for this engine, list the subfolders and load
        # modules
        if os.path.exists(engine_path):
            for filename in os.listdir(engine_path):
                fullpath = os.path.join(engine_path, filename)
                if os.path.isdir(fullpath):
                    # Module names must be unique
                    if filename in modules:
                        raise ModuleError(f"The module {filename} already exists")
                    modules[filename] = engine_class(filename, fullpath)
                    modules_stats.setdefault(engine, 0)
                    modules_stats[engine] += 1
    # Save the modules in `config.MODULES` for reuse by other modules, and print
    # module stats
    config.MODULES = modules
    LOGGER.info(
        "Found %s modules in this package (%s)",
        len(modules),
        ", ".join([f"{number} {engine}" for engine, number in modules_stats.items()]),
    )



"""Module to represent the execution of a package as a graph, with nodes
representing the steps and each step corresponding to a module deployment,
and the edges representing dependencies between steps.
"""

# COMPLETED
import logging
from datetime import datetime, timedelta
from typing import Dict, List, Optional, Tuple

import networkx as nx

from aws_orga_deployer.package.store import ModuleAccountRegionKey

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


class GraphError(Exception):
    """Exception raised when the deployment graph is invalid."""


class NoMorePendingStep(Exception):
    """Exception raised when `next` has no steps to return because all pending
    steps in a graph have been processed.
    """


class NoProcessableStep(Exception):
    """Exception raised when `next` has no steps to return because all remaining
    steps need to wait for their dependencies to complete.
    """


class StepDetails:
    """Store details about a graph step.

    Attributes:
        action: Action to make for this step. Can be "none" (no action
            to make), "create" (deployment to create), "update" (deployment
            to update), "conditional-update" (deployment to update only if the
            values of outputs on which the deployment depends change), "destroy"
            (deployment to destroy).
        skip: True if the step is skipped due to CLI arguments, or because no
            action is needed.
        status: Current step status. Can be "pending", "skipped", "ongoing",
            "completed" or "failed".
        nb_attempts: Number of times this step has been attempted.
        max_attempts: Number of times this step should be attempted before
            it is marked as "failed".
        delay: Delay in seconds to wait before retrying this step.
        wait_until: Must wait until `wait_until` because attempting to process
            this step.
        result: One-line summary of the result.
        detailed_results: Optional detailed results, such as the list of
            resources created, updated or deleted.
        made_changes: True if the step resulted in changes made or to be made
    """

    action: str
    skip: bool
    status: str
    nb_attempts: int
    max_attempts: int
    delay: int
    wait_until: datetime
    result: str
    detailed_results: Optional[Dict]
    made_changes: bool

    def __init__(self, action: str, skip: bool, max_attempts: int, delay: int) -> None:
        self.action = action
        self.skip = skip
        self.status = "pending"
        self.nb_attempts = 0
        self.max_attempts = max_attempts
        self.delay = delay
        self.wait_until = datetime.utcnow()
        self.result = ""
        self.detailed_results = None
        self.made_changes = False


class DeploymentGraph:
    """Class that represents the deployment steps, their dependencies and
    status.
    """

    def __init__(self) -> None:
        self._graph = nx.DiGraph()

    def list_steps(self) -> List[ModuleAccountRegionKey]:
        """Returns the list of steps in the deployment graph.

        Returns:
            List of keys.
        """
        return self._graph.nodes

    def list_steps_with_details(
        self,
    ) -> List[Tuple[ModuleAccountRegionKey, StepDetails]]:
        """Returns the list of steps in the deployment graph and their details.

        Returns:
            List of tuples (key, details).
        """
        return self._graph.nodes(data="details")

    def get_step_details(self, key: ModuleAccountRegionKey) -> StepDetails:
        """Return the details for a given step.

        Args:
            key: Step key

        Returns:
            Step details
        """
        return self._graph.nodes[key]["details"]

    def has_ascendants_with_changes(self, key: ModuleAccountRegionKey) -> bool:
        """Return True if the step has ascendant dependencies on steps with
        pending changes to be made.

        Args:
            key: Step key

        Returns:
            True if at least one ascendant step has changes to be made.
        """
        for ascendant in self._graph.predecessors(key):
            details = self._graph.nodes[ascendant]["details"]
            if details.action != "none" and details.skip is False:
                return True
        return False

    def add_step(
        self,
        key: ModuleAccountRegionKey,
        action: str,
        skip: bool,
        max_attempts: int = 1,
        delay: int = 0,
    ) -> None:
        """Add a new step in the graph, which corresponds to a module
        deployment. Refers to the attributes of `StepDetails` for description.

        Args:
            key: Step key
            action: Action to be made
            skip: True if the step should be skipped
            max_attempts: Maximum number of attemps for this step
            delay: Number of steps to wait before retrying
        """
        details = StepDetails(action, skip, max_attempts, delay)
        self._graph.add_node(key, details=details)

    def add_dependency(
        self,
        from_key: ModuleAccountRegionKey,
        to_key: ModuleAccountRegionKey,
        is_var: bool,
        ignore_if_not_exists: bool = False,
    ) -> None:
        """Add a new edge in the graph, which corresponds to a dependency
        between steps.

        Args:
            from_key: Dependency of the module deployment.
            to_key: Module deployment
            is_var (str): True if the dependency is of type `VariablesFromOutputs`.
            ignore_if_not_exists: Ignore if the dependent deployment does not
                exist if set to True. Otherwise, raise an exception.

        Raises:
            GraphError: If the dependency does not exist in the graph, unless
                the action is `delete` because we don't need to wait for the
                dependency to be deleted if it doesn't exist, or if
                `ignore_if_not_exists` is True.
        """
        to_details = self._graph.nodes[to_key]["details"]
        if not from_key in self._graph.nodes:
            if to_details.action == "destroy" or ignore_if_not_exists is True:
                return
            raise GraphError(f"{to_key} depends on {from_key} which does not exist")
        self._graph.add_edge(from_key, to_key, is_var=is_var)

    def validate(self) -> None:
        """Validate the graph and make initial changes."""
        self._check_for_loops()
        self._check_for_uncreatable_deployments()
        self._check_for_undeletable_deployments()
        self._propagate_conditional_update()
        self._set_status_skipped()
        LOGGER.debug(
            "The deployment graph is valid and contains %s steps and %s dependencies",
            len(self._graph.nodes),
            len(self._graph.edges),
        )

    def _check_for_loops(self) -> None:
        """Check that there are no loops in the graph."""
        if not nx.is_directed_acyclic_graph(self._graph):
            # Identify cycles in the graph
            cycles = []
            for cycle in list(nx.simple_cycles(self._graph)):
                cycles.append(">".join([str(node) for node in cycle]))
            raise GraphError(
                f"The package contains circular dependencies: {' and '.join(cycles)}"
            )

    def _check_for_uncreatable_deployments(self) -> None:
        """Steps that depend on other deployments can only be created if
        their ancestors have been created.
        """
        for to_key, to_details in self._graph.nodes(data="details"):
            if not (to_details.action == "create" and to_details.skip is False):
                continue
            for from_key in self._graph.predecessors(to_key):
                from_details = self._graph.nodes[from_key]["details"]
                if from_details.action == "destroy" and from_details.skip is False:
                    raise GraphError(
                        f"{to_key} must be created after {from_key} which will be"
                        " deleted during this run"
                    )
                if from_details.action == "create" and from_details.skip is True:
                    raise GraphError(
                        f"{to_key} must be created after {from_key} which has not yet"
                        " been created and will not be created during this run"
                    )

    def _check_for_undeletable_deployments(self) -> None:
        """Deployments on which other deployments depend can only be deleted if
        their descendants have been deleted."""
        for from_key, from_details in self._graph.nodes(data="details"):
            if not (from_details.action == "destroy" and from_details.skip is False):
                continue
            for to_key in self._graph.successors(from_key):
                to_details = self._graph.nodes[to_key]["details"]
                if to_details.action == "create" and to_details.skip is True:
                    return
                if to_details.action == "destroy" and to_details.skip is False:
                    return
                raise GraphError(
                    f"{from_key} must be deleted after {to_key} which has not yet"
                    " been deleted and will not be deleted during this run"
                )

    def _propagate_conditional_update(self) -> None:
        """When a deployment A depends on the output values of a deployment B,
        updating B may require updating A if the output values change. Repeat
        until propagation is completed.
        """
        while True:
            changes_made = False
            for from_key, to_key, is_var in self._graph.edges(data="is_var"):
                if is_var:
                    from_details = self._graph.nodes[from_key]["details"]
                    to_details = self._graph.nodes[to_key]["details"]
                    if (
                        from_details.action in ("update", "conditional-update")
                        and to_details.action == "none"
                    ):
                        to_details.action = "conditional-update"
                        changes_made = True
            if not changes_made:
                break

    def _set_status_skipped(self) -> None:
        """Change the status of steps where no actions are needed or steps to
        skip.
        """
        for _, details in self._graph.nodes(data="details"):
            if details.action == "none" or details.skip is True:
                details.status = "skipped"

    def next(self) -> ModuleAccountRegionKey:
        """Return the next step to process.

        Returns:
            The next step key to process.

        Raises:
            aws_orga_deployer.package.graph.NoProcessableStep
            aws_orga_deployer.package.graph.NoMorePendingStep
        """

        def step_is_waiting(details: StepDetails) -> bool:
            """Return True if `wait_until` is in the future."""
            return details.wait_until > datetime.utcnow()

        def mark_as_ongoing(details: StepDetails) -> None:
            """Change the status to "ongoing" and add one to the number of
            attempts.
            """
            details.status = "ongoing"
            details.nb_attempts += 1

        def mark_as_failed(key: ModuleAccountRegionKey, details: StepDetails) -> None:
            """Change the status to "failed" because one of the dependencies
            failed.
            """
            details.status = "failed"
            details.result = "Failed because at least one dependency failed"
            LOGGER.error("%s Failed because at least one dependency failed", key)

        # The topological sort returns the list of steps in the order of which
        # they must be processed (dependencies first)
        path = list(nx.topological_sort(self._graph))
        # We start with delete operations, from the end of the path because
        # deployments must be deleted before their dependencies are deleted
        for key in reversed(path):
            details = self._graph.nodes[key]["details"]
            # Skip if the step is still waiting after a failed attempt
            if step_is_waiting(details):
                continue
            # If a deployment must be deleted
            if details.action == "destroy" and details.status == "pending":
                # If at least one descendant failed, change the status to
                # failed
                if any(
                    self._graph.nodes[descendant_key]["details"].status == "failed"
                    for descendant_key in self._graph.successors(key)
                ):
                    mark_as_failed(key, details)
                # If all descendants have completed, change the status to
                # ongoing and return the node key
                if all(
                    self._graph.nodes[descendant_key]["details"].status
                    in ("completed", "skipped")
                    for descendant_key in self._graph.successors(key)
                ):
                    mark_as_ongoing(details)
                    return key
        # We continue with all create and update operations, from the beginning
        # of the path, if no delete operation was returned
        for key in path:
            details = self._graph.nodes[key]["details"]
            # Skip if the step is still waiting after a failed attempt
            if step_is_waiting(details):
                continue
            # If a node must be created or updated
            if (
                details.action in ("create", "update", "conditional-update")
                and details.status == "pending"
            ):
                # If at least one ancestor failed, the status is changed to
                # failed
                if any(
                    self._graph.nodes[ancestor_key]["details"].status == "failed"
                    for ancestor_key in self._graph.predecessors(key)
                ):
                    mark_as_failed(key, details)
                # If all ancestors have completed or are skipped, the node can
                # be processed
                if all(
                    self._graph.nodes[ancestor_key]["details"].status
                    in ("completed", "skipped")
                    for ancestor_key in self._graph.predecessors(key)
                ):
                    mark_as_ongoing(details)
                    return key
        # Raise NoProcessableStep if there are still pending steps to process
        # but their dependencies must complete first
        if any(
            details.status == "pending"
            for _, details in self._graph.nodes(data="details")
        ):
            raise NoProcessableStep
        # Raise NoMorePendingStep if there is no more steps to process
        raise NoMorePendingStep

    def complete(
        self,
        key: ModuleAccountRegionKey,
        made_changes: bool = False,
        result: str = "",
        detailed_results: Optional[Dict] = None,
    ) -> None:
        """Change the status of a node to "completed" and set the value
        of the result.

        Args:
            key: Step key
            made_changes: True if the step resulted in changes made or to be made
            result: One-line summary of the result. Default to empty.
            detailed_results: Dict with detailed results. Default to None.
        """
        details = self._graph.nodes[key]["details"]
        details.status = "completed"
        details.made_changes = made_changes
        details.result = result
        details.detailed_results = detailed_results

    def fail(
        self,
        key: ModuleAccountRegionKey,
        result: str = "",
        detailed_results: Optional[Dict] = None,
    ) -> None:
        """Change the status of a node to "failed" if no further retries should
        be attempted and set the value of the result.

        Args:
            key: Step key.
            result: One-line summary of the result. Default to empty.
            detailed_results: Dict with detailed results. Default to None.
        """
        details = self._graph.nodes[key]["details"]
        # If the number of maximum attempts is not reached, set the status
        # to "pending"
        details.result = result
        details.detailed_results = detailed_results
        if details.nb_attempts < details.max_attempts:
            details.status = "pending"
            details.wait_until = datetime.utcnow() + timedelta(seconds=details.delay)
        else:
            details.status = "failed"



"""Engine for Python modules."""

from os import path
from typing import Any, Dict, List

from aws_orga_deployer.engines import base
from aws_orga_deployer.package.store import ModuleAccountRegionKey


class Engine(base.BaseEngine):
    """Engine for Python modules."""

    engine = "python"
    default_included_patterns = ["*.py"]
    default_excluded_patterns = []

    def validate_module_config(self, module_config: Dict[str, Any]) -> None:
        """Validate the content of the module configuration.

        Args:
            module_config: Module configuration.

        Raises:
            AssertionError
        """
        super().validate_module_config(module_config)
        if "PythonExecutable" in module_config:
            assert isinstance(
                module_config["PythonExecutable"], str
            ), "PythonExecutable must be a string"

    def prepare(
        self,
        key: ModuleAccountRegionKey,
        command: str,
        action: str,
        variables: Dict[str, Any],
        module_config: Dict[str, Any],
        deployment_cache_dir: str,
        engine_cache_dir: str,
    ) -> List[base.StepCommand]:
        """Create a file `input.json` that contains input parameters, and return
        a command that executes the Python module script in a "wrapper" function.

        Args:
            key: Step key to execute.
            command: CLI command requested ("preview" or "deploy").
            action: Action to be made ("create", "update" or "destroy").
            variables: Input variables.
            module_config: Module configuration.
            deployment_cache_dir: Path to the module deployment cache directory.
            engine_cache_dir: Path to the engine cache directory (common to all
                modules for this engine).

        Returns:
            List of commands to execute in subprocesses.
        """
        # Write the inputs to a JSON file that the wrapper function can read
        self.write_wrapper_inputs(
            key,
            command,
            action,
            variables,
            module_config,
            deployment_cache_dir,
            engine_cache_dir,
        )
        # Locate the Python wrapper script that is executed in a subprocess
        wrapper_file = path.join(
            path.dirname(path.abspath(__file__)), "wrappers", "python.py"
        )
        # Locate the main script of the module
        module_file = path.join(self.module_dir, "main.py")
        # Return a command that execute the wrapper function with the path to
        # the module main script as an argument, and the worker cache folder as
        # the current working directory
        python_exec = base.get_python_executable(module_config)
        return [
            base.StepCommand(
                name="wrapper",
                args=[python_exec, wrapper_file, module_file],
                cwd=deployment_cache_dir,
                assume_role=True,
            )
        ]

    def postprocess(
        self,
        key: ModuleAccountRegionKey,
        command: str,
        action: str,
        module_config: Dict[str, Any],
        deployment_cache_dir: str,
    ) -> base.StepOutcome:
        """Reads the file `output.json` created by the "wrapper" function and
        returns the outputs.

        Args:
            key: Step key to execute.
            command: CLI command requested.
            action: Action to be made ("create", "update", "destroy").
            module_config: Module configuration.
            deployment_cache_dir: Path to the module deployment cache directory.

        Returns:
            Outcome of the step.
        """
        return self.read_wrapper_outputs(deployment_cache_dir)
